from typing import Iterable

import numba
import numpy as np
import pandas as pd
import igraph as ig
import scipy.sparse as sp

import gnnuers.utils as utils


def get_bipartite_igraph(dataset, remove_first_row_col=False):
    inter_matrix = dataset.inter_matrix(form='csr').astype(np.float32)
    if remove_first_row_col:
        inter_matrix = inter_matrix[1:, 1:]

    incid_adj = ig.Graph.Incidence(inter_matrix.todense().tolist())
    bip_info = np.concatenate([np.zeros(inter_matrix.shape[0], dtype=int), np.ones(inter_matrix.shape[1], dtype=int)])

    return ig.Graph.Bipartite(bip_info, incid_adj.get_edgelist())


def _igraph_distances(graph):
    import joblib
    vcount = graph.vcount()
    distances = np.zeros((vcount, vcount), dtype=int)

    def _add_sp(i):
        distances[i] = [len(sp) for sp in graph.get_shortest_paths(i)]

    joblib.Parallel(n_jobs=-1, require='sharedmem')(
        joblib.delayed(_add_sp)(i) for i in range(vcount)
    )

    return distances


class GraphMetricsExtractor():

    BASE_METRICS = ["Degree", "Sparsity", "Reachability", "Sharing Potentiality", "UPI"]

    def __init__(self, dataset, graph_metrics_df=None, metrics="all", sp_kwargs=None, upi_kwargs=None):
        self.dataset = dataset
        self.graph_metrics_df = graph_metrics_df
        self.node_col = 'Node'
        self.metrics = self.BASE_METRICS if metrics == "all" else metrics

        self._item_hist = None
        self._item_hist_len = None

        self._user_hist = None
        self._user_hist_len = None

        self._igg = None

        self.sp_kwargs = sp_kwargs or dict(length=2, depth=2)
        self.upi_kwargs = upi_kwargs or {'sensitive_attribute': ['gender']}

    @property
    def item_hist(self):
        if self._item_hist is None:
            self._item_hist, _, self._item_hist_len = self.dataset.history_user_matrix()
        return self._item_hist

    @property
    def item_hist_len(self):
        if self._item_hist_len is None:
            self._item_hist, _, self._item_hist_len = self.dataset.history_user_matrix()
        return self._item_hist_len

    @property
    def user_hist(self):
        if self._user_hist is None:
            self._user_hist, _, self._user_hist_len = self.dataset.history_item_matrix()
        return self._user_hist

    @property
    def user_hist_len(self):
        if self._user_hist_len is None:
            self._user_hist, _, self._user_hist_len = self.dataset.history_item_matrix()
        return self._user_hist

    @property
    def igg(self):
        if self._igg is None:
            self._igg = get_bipartite_igraph(self.dataset, remove_first_row_col=True)
        return self._igg

    def extract_graph_metrics_per_node(self):
        G_df = self.graph_metrics_df.copy(deep=True)

        for metr in self.metrics:
            if G_df is not None and metr in G_df.columns:
                continue

            if metr == "Degree":
                df = pd.DataFrame(dict(zip(self.igg.vs.indices, self.igg.degree())).items(), columns=[self.node_col, metr])
            elif metr == "Reachability":
                last_user_id = self.dataset.user_num - 2
                user_reach = get_user_reachability(self.igg, last_user_id=last_user_id)

                first_item_id = last_user_id + 1
                item_reach = get_item_reachability(self.igg, first_item_id=first_item_id)

                df = pd.DataFrame({**user_reach, **item_reach}.items(), columns=[self.node_col, metr])
            elif metr == "Sparsity" or metr == "Density":
                _item_hist = self.item_hist[1:].where(self.item_hist[1:] == 0, self.item_hist[1:] - 1)
                item_pop = self.item_hist_len
                _user_hist = self.user_hist[1:].where(self.user_hist[1:] == 0, self.user_hist[1:] - 1)
                user_hist_len = self.user_hist_len

                user_density = np.nan_to_num(
                    ((item_pop[_user_hist] / _user_hist.shape[0]).sum(dim=1) / user_hist_len[1:]).numpy(),
                    nan=0
                )
                user_metric = 1 - user_density if metr == "Sparsity" else user_density

                # item density represents the activity of the users that interact with an item.
                # If only a user interact with item X and this user interacted with all the items in the catalog, then
                # the density of X is maximum. A low density then means a high sparsity, which means the users that interact
                # with that item interact with a few others
                item_density = np.nan_to_num(
                    ((user_hist_len[_item_hist] / _item_hist.shape[0]).sum(dim=1) / item_pop[1:]).numpy(),
                    nan=0
                )
                item_metric = 1 - item_density if metr == "Sparsity" else item_density

                df = pd.DataFrame(
                    zip(igg.vs.indices, np.concatenate([user_metric, item_metric])), columns=[self.node_col, metr]
                )
            elif metr == "Sharing Potentiality":
                n_users = self.user_hist[1:].shape[0]
                user_user = get_node_node_graph_data(self.user_hist[1:].numpy())
                user_user = np.asarray(sp.coo_matrix(
                    (user_user[:, -1], (user_user[:, 0], user_user[:, 1])), shape=(n_users, n_users)
                ).todense())
                user_sp = compute_sharing_potentiality(
                    user_user, self.user_hist_len[1:].numpy(), **self.sp_kwargs
                )

                n_items = self.item_hist[1:].shape[0]
                item_item = get_node_node_graph_data(self.item_hist[1:].numpy())
                item_item = np.asarray(sp.coo_matrix(
                    (item_item[:, -1], (item_item[:, 0], item_item[:, 1])), shape=(n_items, n_items)
                ).todense())
                item_sp = compute_sharing_potentiality(
                    item_item, self.item_hist_len[1:].numpy(), **self.sp_kwargs
                )

                sp_data = np.concatenate([user_sp, item_sp])
                df = pd.DataFrame(zip(range(len(sp_data)), sp_data), columns=[self.node_col, metr])
            elif metr == "Unprotected Preference Index" or metr.upper() == "UPI":
                if self.upi_kwargs is None:
                    raise('UPI needs the upi_kwargs to know the sensitive attribute and the disadvantaged group')
                else:
                    sens_attrs = self.upi_kwargs.pop('sensitive_attribute')
                    sens_attrs = [sens_attrs] if not isinstance(sens_attrs, Iterable) else sens_attrs

                sa_dfs = []
                for sa in sens_attrs:
                    if "UPI " + sa.title() not in G_df.columns:
                        _item_hist = self.item_hist[1:].where(self.item_hist[1:] == 0, self.item_hist[1:] - 1)
                        _user_hist = self.user_hist[1:].where(self.user_hist[1:] == 0, self.user_hist[1:] - 1)

                        sens_map = self.dataset.user_feat[sa]
                        grs_size = sens_map[1:].bincount()
                        minor_gr = ((grs_size > 0) & (grs_size < grs_size.max())).nonzero().item()
                        n_minor = (sens_map == minor_gr).sum() / (sens_map.shape[0] - 1)

                        sens_item_hist = sens_map[_item_hist]
                        items_upi = ((sens_item_hist == minor_gr).sum(dim=1) / self.item_hist_len[1:]).nan_to_num(nan=0)
                        # values higher than 1 means the advantaged group prefer those items w.r.t. to their representation
                        items_upi = items_upi / n_minor

                        users_upi = items_upi[_user_hist].mean(dim=1)

                        upi = np.concatenate((users_upi.numpy(), items_upi.numpy()))
                        sa_dfs.append(pd.DataFrame(zip(range(len(upi)), upi), columns=[self.node_col, "UPI " + sa.title()]).set_index('Node'))
                if not sa_dfs:
                    continue
                df = pd.concat(sa_dfs, axis=1).reset_index()
            if G_df is None:
                G_df = df
            else:
                G_df = G_df.join(df.set_index(self.node_col), on=self.node_col)
        return G_df


def get_user_reachability(graph, last_user_id):
    return get_reachability_per_node(graph, last=last_user_id)


def get_item_reachability(graph, first_item_id):
    return get_reachability_per_node(graph, first=first_item_id)


def get_reachability_per_node(graph, first=None, last=None, nodes=None):
    if nodes is not None:
        nodes = sorted(nodes)
    else:
        nodes = np.arange(graph.vcount())

#         dist = dist[first:][:, first:] if first is not None else dist
#         nodes = nodes[first:] if first is not None else nodes

#         dist = dist[:(last + 1)][:, :(last + 1)] if last is not None else dist
#         nodes = nodes[:(last + 1)] if last is not None else nodes
        if first is not None:
            nodes = nodes[first:]
        if last is not None:
            nodes = nodes[:(last + 1)]

    dist = np.array(graph.distances(source=nodes, target=nodes))

    reach = _get_reachability_per_node(dist)

    return dict(zip(nodes, reach))


def get_reachability_source_target(graph, source, target):
    dist = np.array(graph.distances(source=source, target=target))

    reach = _get_reachability_per_node(dist)

    return dict(zip(nodes, reach))


@numba.jit(nopython=True, parallel=True)
def _get_reachability_per_node(dist):
    n_nodes = dist.shape[0]
    reach = np.zeros((n_nodes,), dtype=np.float32)

    for i in numba.prange(n_nodes):
        n_dist = dist[i]
        mask = (~np.isinf(n_dist)) & (n_dist > 0)
        n_reach = np.bincount(n_dist[mask].astype(np.int32))
        n_reach = n_reach[n_reach > 0]
        reach[i] = compute_reachability(n_reach) / n_nodes

    return reach


@numba.jit(nopython=True, parallel=True)
def compute_reachability(reach):
    discount = np.arange(reach.shape[0]) + 1
    return (reach / discount).sum()


@numba.jit(nopython=True, parallel=True)
def compute_sharing_potentiality(common_data, hist_len, length=2, depth=2):
    n = common_data.shape[0]
    res = np.zeros((n,), dtype=np.float32)
    for i in numba.prange(n):
        most_sim = np.argsort(common_data[i])[::-1][:length]
        res[i] += _compute_sp_length(i, most_sim, common_data, hist_len)
        for d in range(depth - 1):
            most_d_sim = np.argsort(common_data[most_sim[d]])[::-1][:length]
            res[i] += _compute_sp_length(most_sim[d], most_d_sim, common_data, hist_len) / (2 + d)
    return res


@numba.jit(nopython=True)
def _compute_sp_length(data_i, most_sim, common_data, hist_len):
    sp_length = 0
    for i in numba.prange(most_sim.shape[0]):
        sim = common_data[data_i, most_sim[i]]
        if hist_len[data_i] == 0 and hist_len[most_sim[i]] > 0:
            sp_length += 1  # most_sim[i] can share every item
        elif hist_len[data_i] == 0 or hist_len[most_sim[i]] == 0:
            sp_length += 0
        else:
            sp_length += sim / hist_len[data_i] * (1 - sim / hist_len[most_sim[i]])
    return sp_length / most_sim.shape[0]


def get_user_user_data_sens_df(dataset, user_df, sens_attr, attr_map=None):
    user_history, _, _ = dataset.history_item_matrix()
    user_graph_df = get_node_node_data_feature_df(user_history.numpy(), user_df, dataset.uid_field, sens_attr, attr_map=attr_map)

    return user_graph_df


def get_item_item_data_pop_df(dataset, item_df, pop_attr):
    item_history, _, _ = dataset.history_user_matrix()
    item_graph_df = get_node_node_data_feature_df(item_history.numpy(), item_df, dataset.iid_field, pop_attr)

    return item_graph_df


def get_node_node_data_feature_df(history, node_df, id_label, feat_attr, attr_map=None):
    graph = get_node_node_graph_data(history)
    graph = graph[graph[:, -1] > 0]

    graph_df = pd.DataFrame(graph, columns=[f'{id_label}_1', f'{id_label}_2', 'n_common_edges'])
    graph_df = graph_df.join(
        node_df.set_index(id_label), on=f"{id_label}_1"
    ).reset_index(drop=True).rename(columns={feat_attr: f"{feat_attr}_1"})

    if attr_map is not None:
        graph_df[f"{feat_attr}_1"] = graph_df[f"{feat_attr}_1"].map(attr_map)

    graph_df = graph_df.join(
        node_df.set_index(id_label), on=f'{id_label}_2'
    ).reset_index(drop=True).rename(columns={feat_attr: f"{feat_attr}_2"})

    if attr_map is not None:
        graph_df[f"{feat_attr}_2"] = graph_df[f"{feat_attr}_2"].map(attr_map)

    return graph_df


def get_node_node_graph_data(history):
    return _get_node_node_graph_data(history)


@numba.jit(nopython=True, parallel=True)
def _get_node_node_graph_data(history):
    hist = [set(h) for h in history]
    n_nodes = history.shape[0] - 1  # removed padding 0
    node_node = np.zeros((n_nodes * (n_nodes - 1) // 2, 3), dtype=np.int32)  # number of combinations

    for n1 in numba.prange(1, n_nodes + 1):
        _inner_combinations(n1, n_nodes, hist, node_node)

    return node_node


@numba.jit(nopython=True, parallel=True)
def _inner_combinations(n1, n_nodes, hist, node_node):
    for n2 in numba.prange(n1 + 1, n_nodes + 1):
        node_node[sum(range(n_nodes - n1 + 1, n_nodes)) + (n2 - n1 - 1)] = [
            n1, n2, len(hist[n1] & hist[n2]) - 1
        ]


def compute_homophily(graph_df, group_sizes, feat_attr):
    homophily = {}
    for (gr1, gr2), gr_df in graph_df.groupby([f"{feat_attr}_1", f"{feat_attr}_2"]):
        if gr1 == gr2:
            all_gr_edges = graph_df.loc[
                (graph_df[f"{feat_attr}_1"] == gr1) |  # interactions of gr1 with any group (also gr1)
                ((graph_df[f"{feat_attr}_2"] == gr1) & (graph_df[f"{feat_attr}_1"] != gr1)),  # interactions from gr2 to gr1
                'n_common_edges'
            ].sum()
            homophily[gr1] = gr_df['n_common_edges'].sum() / all_gr_edges - group_sizes[gr1]

    return homophily



import os
import re
import json
import yaml
import pickle
import logging
import inspect
import argparse

import wandb
import torch
import optuna
import pandas as pd

from recbole.data.dataloader import FullSortEvalDataLoader

import gnnuers.utils as utils
from gnnuers.explainers import DPBG, BaB


script_path = os.path.abspath(os.path.dirname(inspect.getsourcefile(lambda: 0)))


def get_base_exps_filepath(config,
                           config_id=-1,
                           model_name=None,
                           model_file="",
                           hyper=False):
    """
    return the filepath where explanations are saved
    :param config:
    :param config_id:
    :param model_name:
    :param model_file:
    :return:
    """
    epochs = config["cf_epochs"]
    model_name = model_name or config["model"]
    explainer = config["explainer"].lower()
    exp_type = "dp_explanations" if not hyper else "hyperoptimization"
    base_exps_file = os.path.join(script_path, "experiments", exp_type, config["dataset"], model_name, explainer)

    exp_metadata = config["sensitive_attribute"].lower() if 'consumer' in config["exp_metric"].lower() else ''
    exp_loss = config["exp_metric"] + "_loss"
    base_exps_file = os.path.join(base_exps_file, exp_loss, exp_metadata, f"epochs_{epochs}")

    if os.path.exists(base_exps_file):
        if config_id == -1:
            paths_c_ids = sorted(filter(str.isdigit, os.listdir(base_exps_file)), key=int)
            if len(paths_c_ids) == 0:
                config_id = 1
            else:
                int_paths_c_ids = list(map(int, paths_c_ids))
                candidates = set(range(1, max(int_paths_c_ids) + 1)) - set(int_paths_c_ids)
                config_id = str(min(candidates, default=max(int_paths_c_ids) + 1))

            for path_c in paths_c_ids:
                config_path = os.path.join(base_exps_file, path_c, "config.pkl")
                if os.path.exists(config_path):
                    with open(config_path, 'rb') as f:
                        _c = pickle.load(f)

                    if config.final_config_dict == _c.final_config_dict:
                        if model_file != "" and "perturbed" in model_file:
                            check_perturb = input("The explanations of the perturbed graph could overwrite the "
                                                  "explanations from which the perturbed graph was generated. Type "
                                                  "y/yes to confirm this outcome. Other inputs will assign a new id: ")
                            if check_perturb.lower() != "y" and check_perturb.lower() != "yes":
                                continue
                        config_id = os.path.join(base_exps_file, str(path_c))
                        break
                elif hyper:
                    config_id = os.path.join(base_exps_file, str(path_c))
                    break


        base_exps_file = os.path.join(base_exps_file, str(config_id))
    else:
        base_exps_file = os.path.join(base_exps_file, "1")

    if not os.path.exists(base_exps_file):
        os.makedirs(base_exps_file)

    with open(os.path.join(base_exps_file, "config.yaml"), 'w') as exp_file:
        exp_file.write(yaml.dump(config.final_config_dict, default_flow_style=False))

    return base_exps_file


def save_exps_df(base_exps_file, exps):
    """
    Saves the pandas dataframe representation of explanations
    :param base_exps_file:
    :param exps:
    :return:
    """
    data = [exp[:-3] + exp[-2:] for exp_list in exps.values() for exp in exp_list]
    data = list(map(lambda x: [x[0], *x[1:]], data))
    df = pd.DataFrame(
        data,
        columns=utils.EXPS_COLUMNS[:-3] + utils.EXPS_COLUMNS[-2:]
    )

    out_path = base_exps_file.split(os.sep)
    df.to_csv(
        f'{"_".join(out_path[out_path.index("explanations"):])}.csv',
        index=False
    )


def explain(config, model, _rec_data, _full_dataset, _train_data, _valid_data, _test_data, base_exps_file, **kwargs):
    """
    Function that explains, that is generates perturbed graphs.
    :param config:
    :param model:
    :param _train_dataset:
    :param _rec_data:
    :param _test_data:
    :param base_exps_file:
    :param kwargs:
    :return:
    """
    epochs = config['cf_epochs']
    topk = config['cf_topk']
    wandb_mode = kwargs.get("wandb_mode", "disabled")
    overwrite = kwargs.get("overwrite", False)

    exps_filename = os.path.join(base_exps_file, f"cf_data.pkl")
    users_order_file = os.path.join(base_exps_file, f"users_order.pkl")
    model_preds_file = os.path.join(base_exps_file, f"model_rec_test_preds.pkl")
    checkpoint_path = os.path.join(base_exps_file, "checkpoint.pth")

    if overwrite and os.path.exists(checkpoint_path):
        os.remove(checkpoint_path)

    user_source = _rec_data if _rec_data is not None else _test_data
    user_data = user_source.user_df[user_source.uid_field][torch.randperm(user_source.user_df.length)]

    with open(os.path.join(base_exps_file, "config.pkl"), 'wb') as config_file:
        pickle.dump(config, config_file)

    utils.wandb_init(
        config,
        **wandb_env_data,
        name="Explanation",
        job_type="train",
        group=f"{model.__class__.__name__}_{config['dataset']}_{config['sensitive_attribute'].title()}_epochs{config['cf_epochs']}_exp{os.path.basename(base_exps_file)}",
        mode=wandb_mode
    )
    # wandb.config.update({"exp": os.path.basename(base_exps_file)})

    explainer_model = {
        "dpbg": DPBG,
        "bab": BaB
    }.get(config["explainer"].lower(), DPBG)

    explainer = explainer_model(config, _train_data.dataset, _rec_data, model, dist=config['cf_dist'], **kwargs)
    explainer.set_checkpoint_path(checkpoint_path)
    exp, users_order, model_preds = explainer.explain(user_data, _full_dataset, _train_data, _valid_data, _test_data, epochs, topk=topk)

    with open(exps_filename, 'wb') as f:
        pickle.dump(exp, f)
    with open(users_order_file, 'wb') as f:
        pickle.dump(users_order, f)
    with open(model_preds_file, 'wb') as f:
        pickle.dump(model_preds, f)

    logging.getLogger().info(f"Saved explanations at path {base_exps_file}")


def optimize_explain(config, model, _train_dataset, _rec_data, _test_data, base_exps_file, **kwargs):
    """
    Function that explains, that is generates perturbed graphs.
    :param config:
    :param model:
    :param _train_dataset:
    :param _rec_data:
    :param _test_data:
    :param base_exps_file:
    :param kwargs:
    :return:
    """
    epochs = config['cf_epochs']
    topk = config['cf_topk']
    wandb_mode = kwargs.get("wandb_mode", "disabled")

    user_source = _rec_data if _rec_data is not None else _test_data
    user_data = user_source.user_df[user_source.uid_field][torch.randperm(user_source.user_df.length)]

    explainer_model = {
        "dpbg": DPBG,
        "bab": BaB
    }.get(config["explainer"].lower(), DPBG)

    exp_token = f"{model.__class__.__name__}_" + \
                f"{config['dataset']}_" + \
                f"{config['sensitive_attribute'].title()}_" + \
                f"epochs{config['cf_epochs']}_" + \
                f"exp{os.path.basename(base_exps_file)}"

    def objective(trial):
        wandb_config_keys = [
            'cf_learning_rate',
            'user_batch_exp',
            'cf_beta',
            'dropout_prob'
        ]

        config['cf_learning_rate'] = trial.suggest_int('cf_learning_rate', 1000, 10000)

        config['user_batch_exp'] = trial.suggest_int(
            'user_batch_exp',
            min(int(_test_data.dataset.user_num * 0.1), 32),
            min(int(_test_data.dataset.user_num * 0.33), 220)
        )

        config['cf_beta'] = trial.suggest_float('cf_beta', 0.01, 10.0)

        # config['dropout_prob'] = trial.suggest_float('dropout_prob', 0, 0.3)

        if config["explainer"].lower() == "bab":
            config['bab_min_del_edges'] = trial.suggest_int('bab_min_del_edges', 10, 200)
            config['bab_max_tries'] = trial.suggest_int('bab_max_tries', 50, 400)
            wandb_config_keys.extend(['bab_min_del_edges', 'bab_max_tries'])

        wandb_config = {k: config[k] for k in wandb_config_keys}

        run = utils.wandb_init(
            wandb_config,
            **wandb_env_data,
            policies=config['explainer_policies'],
            name=f"Explanation_trial{trial.number}",
            job_type="train",
            group=exp_token,
            mode=wandb_mode,
            reinit=True
        )

        explainer = explainer_model(config, _train_dataset, _rec_data, model, dist=config['cf_dist'], **kwargs)
        exp, model_preds = explainer.explain(user_data, _test_data, epochs, topk=topk)
        best_exp = utils.get_best_exp_early_stopping(exp, config)

        exp_metric = best_exp[utils.exp_col_index('exp_metric')]

        with run:
            run.log({'trial_exp_metric': exp_metric})

        return exp_metric

    study_name = exp_token + '_' + str([k for k in config['explainer_policies'] if config['explainer_policies'][k]])
    storage_name = "sqlite:///{}.db".format(study_name)
    study = optuna.create_study(direction="minimize", study_name=study_name, storage=storage_name, load_if_exists=True)

    n_trials = 100
    if study.trials:
        n_trials -= study.trials[-1].number  # it is not done automatically by optuna
        if n_trials <= 0:
            raise ValueError(f"Optuna study with storage name {study_name}.db is already completed")

    study.optimize(objective, n_trials=n_trials)

    summary = utils.wandb_init(
        config,
        **wandb_env_data,
        name="summary",
        job_type="logging",
        group=exp_token,
        mode=wandb_mode
    )

    trials = study.trials

    print("Number of finished trials: ", len(trials))

    # WandB summary.
    for step, trial in enumerate(trials):
        # Logging the loss.
        summary.log({"trial_exp_metric": trial.value}, step=step)

        # Logging the parameters.
        for k, v in trial.params.items():
            summary.log({k: v}, step=step)

    with open(os.path.join(base_exps_file, 'best_params.json'), 'w') as param_file:
        json.dump(dict(trial.params.items()), param_file, indent=4)


def execute_explanation(model_file,
                        base_explainer_config_file,
                        explainer_config_file=os.path.join("config", "explainer.yaml"),
                        config_id=-1,
                        verbose=False,
                        wandb_mode="disabled",
                        cmd_config_args=None,
                        hyperoptimization=False,
                        overwrite=False):
    explainer_config = utils.update_base_explainer(base_explainer_config_file, explainer_config_file)

    # load trained model, config, dataset
    config, model, dataset, train_data, valid_data, test_data = utils.load_data_and_model(
        model_file,
        explainer_config,
        cmd_config_args=cmd_config_args
    )

    # force these evaluation metrics to be ready to be computed
    config['metrics'] = ['ndcg', 'recall', 'hit', 'mrr', 'precision']

    if config['exp_rec_data'] is not None:
        if config['exp_rec_data'] != 'train+valid':
            if config['exp_rec_data'] == 'train':
                rec_data = FullSortEvalDataLoader(config, train_data.dataset, train_data.sampler)
            elif config['exp_rec_data'] == 'rec':  # model recommendations are used as target
                rec_data = valid_data
            else:
                rec_data = locals()[f"{config['exp_rec_data']}_data"]
        else:
            valid_train_dataset = train_data.dataset.copy(
                pd.concat([train_data.dataset.inter_feat, valid_data.dataset.inter_feat], ignore_index=True)
            )
            rec_data = FullSortEvalDataLoader(config, valid_train_dataset, valid_data.sampler)
    else:
        rec_data = valid_data

    base_exps_filepath = get_base_exps_filepath(
        config,
        config_id=config_id,
        model_name=model.__class__.__name__,
        model_file=model_file,
        hyper=hyperoptimization
    )

    if not os.path.exists(base_exps_filepath):
        os.makedirs(base_exps_filepath)

    global wandb_env_data
    wandb_env_data = {}
    if os.path.exists("wandb_init.json"):
        with open("wandb_init.json", 'r') as wandb_file:
            wandb_env_data = json.load(wandb_file)

    kwargs = dict(
        verbose=verbose,
        wandb_mode=wandb_mode,
        overwrite=overwrite
    )

    if not hyperoptimization:
        explain(
            config,
            model,
            rec_data,
            dataset,
            train_data,
            valid_data,
            test_data,
            base_exps_filepath,
            **kwargs
        )
    else:
        optimize_explain(
            config,
            model,
            train_data.dataset,
            rec_data,
            test_data,
            base_exps_filepath,
            **kwargs
        )



import time

import tqdm
import torch
import numpy as np
import pandas as pd
from recbole.utils import set_color

import gnnuers.evaluation as eval_utils

from .base import Explainer


class BaB(Explainer):
    """
    BaB: Batch after Batch
    """
    def __init__(self, config, dataset, rec_data, model, dist="damerau_levenshtein", **kwargs):
        super(BaB, self).__init__(config, dataset, rec_data, model, dist=dist, **kwargs)
        self.mini_batch_descent = None
        self.min_del_edges_batch = config['bab_min_del_edges'] or 1
        self.max_tries_batch = config['bab_max_tries'] or 20

    def compute_rec_test_eval_metric_model_and_cf(self, new_example, test_data, test_model_topk, cf_test_topk):
        test_pref = pd.DataFrame(
            zip(new_example[0], test_model_topk, cf_test_topk),
            columns=['user_id', 'topk_pred', 'cf_topk_pred']
        )
        orig_test_res = eval_utils.compute_metric(
            self.evaluator, test_data, test_pref, 'topk_pred', self.eval_metric
        )
        cf_test_res = eval_utils.compute_metric(
            self.evaluator, test_data, test_pref, 'cf_topk_pred', self.eval_metric
        )

        rec_pref = pd.DataFrame(
            zip(*new_example[:3]), columns=['user_id', 'topk_pred', 'cf_topk_pred']
        )
        orig_rec_res = self.compute_eval_metric(self.rec_data.dataset, rec_pref, 'topk_pred')
        cf_rec_res = self.compute_eval_metric(self.rec_data.dataset, rec_pref, 'cf_topk_pred')

        return orig_test_res, cf_test_res, orig_rec_res, cf_rec_res

    def run_epoch(self, epoch, batched_data, test_data, fair_losses, topk=10):
        iter_data = self.prepare_iter_batched_data(batched_data)
        iter_data = [iter_data[i] for i in np.random.permutation(len(iter_data))]

        bab_epoch_data = []
        epoch_fair_loss = []
        for batch_idx, batch_user in enumerate(iter_data):
            new_example, loss_total = None, None
            batch_scores_args, _ = self._get_model_score_data(batch_user, self.rec_data, topk)

            def batch_check(ne, tc):
                return ne is None or (ne[-2].shape[1] < self.min_del_edges_batch and tc < self.max_tries_batch)

            tries_counter = 0
            del_edges_batch_first_iter = False

            while batch_check(new_example, tries_counter):
                torch.cuda.empty_cache()
                new_example, loss_total, fair_loss = self.train(
                    epoch, batch_scores_args, batch_user, topk=topk, previous_loss_value=None
                )

                if batch_check(new_example, tries_counter):
                    loss_total.backward()
                    torch.nn.utils.clip_grad_norm_(self.cf_model.parameters(), 2.0)
                    self.cf_optimizer.step()

                    del_edges_batch_first_iter = True
                tries_counter += 1

            if del_edges_batch_first_iter:
                self.cf_model.reset_param()

            epoch_fair_loss.append(fair_loss)

            if new_example is not None:
                uids = new_example[0]
                items_involved = np.asarray(np.unique(new_example[-2][1], return_counts=True))

                batch_test_scores_args, test_model_topk_idx = self._get_model_score_data(uids, test_data, topk)
                cf_test_topk_pred_idx, _ = self._get_no_grad_pred_model_score_data(batch_test_scores_args)

                orig_test_res, cf_test_res, orig_rec_res, cf_rec_res = self.compute_rec_test_eval_metric_model_and_cf(
                    new_example, test_data.dataset, test_model_topk_idx, cf_test_topk_pred_idx
                )

                fair_metric_data = []
                sens_attr = self.sensitive_attribute
                for res, dset in zip(
                    [orig_rec_res, orig_test_res, cf_rec_res, cf_test_res],
                    [self.rec_data.dataset, test_data.dataset, self.rec_data.dataset, test_data.dataset]
                ):
                    gr1_mask = dset.user_feat[sens_attr][uids].numpy() == self.adv_group
                    gr2_mask = dset.user_feat[sens_attr][uids].numpy() == self.disadv_group
                    fair_metric_data.append(eval_utils.compute_dp_with_masks(res[:, -1], gr1_mask, gr2_mask))

                bab_epoch_data.append([
                    new_example[0],
                    new_example[-2],
                    items_involved,
                    np.stack((orig_rec_res[:, -1], orig_test_res[:, -1])),
                    np.stack((cf_rec_res[:, -1], cf_test_res[:, -1])),
                    fair_metric_data[:2],
                    fair_metric_data[2:]
                ])
            else:
                bab_epoch_data.append(None)

        epoch_fair_loss = np.mean(epoch_fair_loss)
        fair_losses.append(epoch_fair_loss)

        return bab_epoch_data, epoch_fair_loss

    def explain(self, batched_data, test_data, epochs, topk=10):
        """
        The method from which starts the perturbation of the graph by optimization of `pred_loss` or `fair_loss`
        :param batched_data:
        :param test_data:
        :param epochs:
        :param topk:
        :return:
        """
        best_cf_example = []

        self._check_loss_trend_epoch_images()

        test_scores_args, test_model_topk = self._get_model_score_data(batched_data, test_data, topk)

        # recommendations generated by the model are considered the ground truth
        if self._pred_as_rec:
            rec_scores_args, rec_model_topk = self._prepare_test_history_matrix(test_data, topk=topk)
        else:
            rec_scores_args, rec_model_topk = self._get_model_score_data(batched_data, self.rec_data, topk)

        batched_data, filtered_users, inc_disp_model_data = self._check_policies(batched_data, rec_model_topk)
        if self.increase_disparity:
            test_model_topk, test_scores_args, rec_model_topk, rec_scores_args = inc_disp_model_data

        # logs of fairness consider validation as seen when model recommendations are used as ground truth
        if self._pred_as_rec:
            self.rec_data = test_data

        self.initialize_cf_model(filtered_users=filtered_users)

        orig_loss = np.abs(self.results[self.adv_group] - self.results[self.disadv_group])
        self.logger.info(self.results)
        self.logger.info(f"M idx: {self.m_idx}")
        self.logger.info(f"Original Fair Loss: {orig_loss}")

        iter_epochs = tqdm.tqdm(
            range(epochs),
            total=epochs,
            ncols=100,
            desc=set_color(f"Epochs   ", 'blue'),
        )

        bab_data = []
        fair_losses = []
        for epoch in iter_epochs:
            bab_epoch_data, epoch_fair_loss = self.run_epoch(
                epoch, batched_data, test_data, fair_losses, topk=topk
            )

            if self.verbose:
                self._verbose_plot(fair_losses, epoch)

            bab_data.append(bab_epoch_data)

            # if new_example is not None:
            #     self.update_new_example(
            #         new_example,
            #         detached_batched_data,
            #         test_scores_args,
            #         test_model_topk,
            #         rec_scores_args,
            #         rec_model_topk,
            #         epoch_fair_loss
            #     )
            #
            #     update_best_example_args = [best_cf_example, new_example, loss_total, best_loss, orig_loss]
            #     # if self._check_early_stopping(epoch_fair_metric, *update_best_example_args)
            #     if self._check_early_stopping(epoch_fair_loss, *update_best_example_args):
            #         break
            #
            #     best_loss = self.update_best_cf_example(*update_best_example_args, model_topk=rec_model_topk)

            self.logger.info("{} CF examples".format(len(best_cf_example)))

        return bab_data, (rec_model_topk, test_model_topk)

    def train(self, epoch, scores_args, users_ids, topk=10, previous_loss_value=None):
        t = time.time()

        # `_get_no_grad_pred_model_score_data` call torch `eval` inside
        cf_topk_pred_idx, _ = self._get_no_grad_pred_model_score_data(scores_args)

        self.cf_optimizer.zero_grad()
        self.cf_model.train()

        # compute differentiable permutation of adj matrix
        # cf_scores uses differentiable P_hat ---> adjacency matrix not binary, but needed for training
        cf_scores = self.get_scores(self.cf_model, *scores_args, pred=False)

        # remove neginf from output
        cf_scores = torch.nan_to_num(cf_scores, neginf=(torch.min(cf_scores[~torch.isinf(cf_scores)]) - 1).item())
        cf_scores_topk, cf_topk_idx = self.get_top_k(cf_scores, **self.topk_args)

        user_feat = self.get_batch_user_feat(users_ids)

        target = self.get_target(cf_scores, user_feat)

        loss_total, orig_loss_graph_dist, loss_graph_dist, fair_loss, adj_sub_cf_adj = self.cf_model.loss(
            cf_scores,
            self._fair_loss(
                self.sensitive_attribute,
                user_feat,
                topk=topk,
                loss=self._metric_loss,
                adv_group_data=(self.only_adv_group, self.disadv_group, self.results[self.disadv_group]),
                previous_loss_value=previous_loss_value
            ),
            target
        )

        torch.cuda.empty_cache()
        self.cf_optimizer.zero_grad()

        # import pdb; pdb.set_trace()
        # for name, param in self.cf_model.named_parameters():
        #     if name == "P_symm":
        #         print(param.grad[param.grad])

        fair_loss = fair_loss.mean().item() if fair_loss is not None else torch.nan
        self.log_epoch(
            t, epoch, loss_total, fair_loss, loss_graph_dist, orig_loss_graph_dist,
            **dict(cf_topk_idx=cf_topk_idx, cf_topk_pred_idx=cf_topk_pred_idx)
        )

        cf_stats = None
        if orig_loss_graph_dist.item() > 0:
            cf_stats = self.get_batch_cf_stats(
                users_ids, adj_sub_cf_adj, cf_topk_pred_idx, loss_total, loss_graph_dist, fair_loss, epoch
            )

        return cf_stats, loss_total, fair_loss



import abc

import torch


class RankingLoss(torch.nn.modules.loss._Loss, metaclass=abc.ABCMeta):
    __constants__ = ['reduction']

    def __init__(self, size_average=None, reduce=None, reduction: str = 'mean', temperature=0.01, **kwargs) -> None:
        super(RankingLoss, self).__init__(size_average, reduce, reduction)
        self.temperature = temperature

    @abc.abstractmethod
    def forward(self, _input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("subclasses must implement this method")


class TopKLoss(RankingLoss):
    def __init__(self, size_average=None, reduce=None, topk=None, reduction: str = 'mean', temperature=0.01) -> None:
        super(TopKLoss, self).__init__(
            size_average=size_average,
            reduce=reduce,
            reduction=reduction,
            temperature=temperature
        )
        self.topk = topk

    def forward(self, _input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError("subclasses must implement this method")


class NDCGApproxLoss(TopKLoss):
    __MAX_TOPK_ITEMS__ = 10000

    def __init__(self, size_average=None, reduce=None, topk=None, reduction: str = 'mean', temperature=0.01) -> None:
        """
        Lower values of `temperature` makes the loss more accurate in approximating NDCG
        """

        super(NDCGApproxLoss, self).__init__(
            size_average=size_average,
            reduce=reduce,
            reduction=reduction,
            topk=topk,
            temperature=temperature
        )

    def forward(self, _input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        if _input.shape[1] > self.__MAX_TOPK_ITEMS__:
            topk = self.topk or target.shape[1]
            _, _input_topk = torch.topk(_input, dim=1, k=topk)
            _input = _input[torch.arange(_input.shape[0])[:, None], _input_topk]
            target = target[torch.arange(target.shape[0])[:, None], _input_topk]

        _input_temp = torch.nn.ReLU()(_input) / self.temperature

        def approx_ranks(inp):
            shape = inp.shape[1]

            a = torch.tile(torch.unsqueeze(inp, 2), [1, 1, shape])
            a = torch.transpose(a, 1, 2) - a
            return torch.sum(torch.sigmoid(a), dim=-1) + .5

        def inverse_max_dcg(_target,
                            gain_fn=lambda _target: torch.pow(2.0, _target) - 1.,
                            rank_discount_fn=lambda rank: 1. / rank.log1p()):
            topk = self.topk or _target.shape[1]
            ideal_sorted_target = torch.topk(_target, topk).values
            rank = (torch.arange(ideal_sorted_target.shape[1]) + 1).to(_target.device)
            discounted_gain = gain_fn(ideal_sorted_target).to(_target.device) * rank_discount_fn(rank)
            discounted_gain = torch.sum(discounted_gain, dim=1, keepdim=True)
            return torch.where(discounted_gain > 0., 1. / discounted_gain, torch.zeros_like(discounted_gain))

        def ndcg(_target, _ranks):
            topk = self.topk or _target.shape[1]
            sorted_target, sorted_idxs = torch.topk(_target, topk)
            discounts = 1. / _ranks[torch.arange(_ranks.shape[0])[:, None], sorted_idxs].log1p()
            gains = torch.pow(2., sorted_target).to(_target.device) - 1.
            dcg = torch.sum(gains * discounts, dim=-1, keepdim=True)
            return dcg * inverse_max_dcg(_target)

        ranks = approx_ranks(_input_temp)

        return -ndcg(target, ranks)


class SigmoidBCELoss(RankingLoss):
    """"
    Based on TensorFlow Ranking SigmoidBCELoss
    """

    def __init__(self, size_average=None, reduce=None, topk=None, reduction: str = 'mean', temperature=0.1) -> None:
        super(SigmoidBCELoss, self).__init__(
            size_average=size_average,
            reduce=reduce,
            reduction=reduction,
            topk=topk,
            temperature=temperature
        )

    def forward(self, _input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        _input_temp = _input / self.temperature

        target_sum = target.sum(dim=1, keepdim=True)
        padded_target = torch.where(target_sum > 0, target, torch.ones_like(target) * 1e-7)
        padded_target_sum = padded_target.sum(dim=1, keepdim=True)
        target = torch.nan_to_num(padded_target / padded_target_sum, nan=0)

        return torch.nn.BCEWithLogitsLoss(reduction='none')(_input_temp, target).mean(dim=1, keepdim=True)



import os
import pickle
import logging

import torch
import numpy as np
import pandas as pd
from recbole.utils import FeatureType, FeatureSource, set_color

import gnnuers.utils as utils
from gnnuers.data import Dataset


class PerturbedDataset(Dataset):
    FEATS = [FeatureSource.INTERACTION, FeatureSource.ITEM]  # our method never deletes users

    def __init__(self, config, explanations_path, best_exp):
        self.explanations_path = explanations_path
        self.__base_dataset = Dataset(config)  # must be instantiated to get the recbole mappings of user and item ids

        self.best_exp = best_exp
        self.best_explanation = None
        self.exps_config = None
        self.perturbed_edges = self.load_perturbed_edges()
        self.mapped_perturbed_edges = utils.remap_edges_recbole_ids(
            self.__base_dataset, self.perturbed_edges, field2id_token=True
        )

        super(PerturbedDataset, self).__init__(config)

    def get_best_explanation(self, exps, exps_config):
        if self.best_exp[0] == "fairest":
            best_exp = utils.get_best_exp_early_stopping(exps, exps_config)
        elif self.best_exp[0] == "fixed_exp":
            best_exp = utils.get_exp_by_epoch(exps, self.best_exp[1])
        else:
            raise ValueError("`best_exp` must be set to select the explanation of a specific epoch")

        return best_exp

    def load_perturbed_edges(self):
        logger = logging.getLogger()

        with open(os.path.join(self.explanations_path, 'cf_data.pkl'), 'rb') as exp_file:
            exps = pickle.load(exp_file)
        logger.info(f"Original Fair Loss: {exps[0][-1]}")

        self.exps_config = utils.read_recbole_config_skip_errors(
            os.path.join(self.explanations_path, 'config.yaml'),
            self.__base_dataset.config
        )
        logger.info(self.exps_config)

        if self.exps_config['exp_rec_data'] != 'valid':
            logger.warning('Performing Graph Augmentation on Explanations NOT produced on Validation Data.')

        self.best_explanation = self.get_best_explanation(exps, self.exps_config)

        return self.best_explanation[utils.exp_col_index('del_edges')]

    def perturb_split(self, split):
        if split in self.SPLITS:
            spl_file = os.path.join(self.config['data_path'], f"{self.config['dataset']}.{split}")
            if not os.path.isfile(spl_file):
                return None

            with open(spl_file, 'rb') as split_data_file:
                split_data = pickle.load(split_data_file)

            if isinstance(split_data[self.uid_field], torch.Tensor):
                split_data = {k: v.numpy() for k, v in split_data.items()}

            # Recbole tokens are always treated as strings (e.g., user\item ids)
            split_data = {k: v.astype(str) for k, v in split_data.items()}

            mapped_split_data = utils.remap_edges_recbole_ids(
                self.__base_dataset,
                np.stack((split_data[self.uid_field], split_data[self.iid_field])),
                field2id_token=False
            )

            new_split_data, unique, counts = utils.np_unique_cat_recbole_interaction(
                mapped_split_data, self.perturbed_edges,
                uid_field=self.uid_field, iid_field=self.iid_field, return_unique_counts=True
            )

            if split != "train":
                # validation and test set should not be affected by any deleted edge in the training set
                # but if an edge is added to the training set, it must be removed from the validation or test set
                common_interactions = unique[:, counts > 1]
                new_split_data = utils.np_unique_cat_recbole_interaction(
                    mapped_split_data, common_interactions, uid_field=self.uid_field, iid_field=self.iid_field
                )

            new_split_data = utils.remap_edges_recbole_ids(
                self.__base_dataset,
                np.stack((new_split_data[self.uid_field], new_split_data[self.iid_field])),
                field2id_token=True
            )

            return dict(zip([self.uid_field, self.iid_field], new_split_data))
        else:
            raise ValueError(f"split `{split}` is not supported for LRS configuration")

    def perturb_feat(self, source=FeatureSource.INTERACTION):
        if source in self.FEATS:
            feat_file = os.path.join(self.config['data_path'], f"{self.config['dataset']}.{source.value}")
            if not os.path.isfile(feat_file):
                return None  # it could be triggered if the dataset does not have item features (i.e. ".item" file)

            # new dataset is the merge of splits => it takes care of particular cases, like an edge in the validation
            # set that was added to the training set (it should then not be considered as deleted, and it should be in
            # the final dataset
            pert_splits = [self.perturb_split(split) for split in self.SPLITS]
            merged_splits = pert_splits[0]
            for p_split in pert_splits[1:]:
                for key in merged_splits:
                    merged_splits[key] = np.concatenate([merged_splits[key], p_split[key]])

            # Recbole tokens are always treated as strings (e.g., user\item ids)
            pert_df = pd.DataFrame(merged_splits).astype(str)

            feat_df_cols = [self.uid_field, self.iid_field]
            if source == FeatureSource.INTERACTION:
                feat_df = self.__base_dataset.inter_feat
            elif source == FeatureSource.ITEM and self.__base_dataset.item_feat is not None:
                feat_df = self.__base_dataset.item_feat
                pert_df = pert_df[[self.iid_field]]
            else:
                raise ValueError(f"feat {source} not supported for modification when re-training with perturbed data")

            for field, token_map in self.__base_dataset.field2id_token.items():
                if field in feat_df.columns:
                    # Recbole tokens are always treated as strings (e.g., user\item ids)
                    feat_df[field] = feat_df[field].map(token_map.__getitem__).astype(str)
            new_feat_df = pert_df.join(feat_df.set_index(feat_df_cols), on=feat_df_cols, how="left")

            return new_feat_df
        else:
            raise ValueError(f"feat {source} not supported for modification when re-training with perturbed data")

    def _remap(self, remap_list):
        tokens, split_point = self._concat_remaped_tokens(remap_list)

        common_field = remap_list[0][1]
        new_ids_list = [self.__base_dataset.field2token_id[common_field][orig_id] for orig_id in tokens]

        new_ids_list = np.split(new_ids_list, split_point)

        mp = self.__base_dataset.field2id_token[common_field]
        token_id = self.__base_dataset.field2token_id[common_field]

        for (feat, field, ftype), new_ids in zip(remap_list, new_ids_list):
            if field not in self.field2id_token:
                self.field2id_token[field] = mp
                self.field2token_id[field] = token_id
            if ftype == FeatureType.TOKEN:
                feat[field] = new_ids
            elif ftype == FeatureType.TOKEN_SEQ:
                split_point = np.cumsum(feat[field].agg(len))[:-1]
                feat[field] = np.split(new_ids, split_point)

    def _load_data_split(self, split):
        filename = os.path.join(self.dataset_path, f'{self.dataset_name}.{split}')
        if not os.path.isfile(filename):
            if split in ['train', 'test']:
                raise NotImplementedError(f'The splitting method "LRS" needs at least train and test.')

        split_data = self.perturb_split(split)

        return split_data

    def _load_feat(self, filepath, source):
        """Load features according to source into :class:`pandas.DataFrame`.

        Set features' properties, e.g. type, source and length.

        Args:
            filepath (str): path of input file.
            source (FeatureSource or str): source of input file.

        Returns:
            pandas.DataFrame: Loaded feature

        Note:
            For sequence features, ``seqlen`` will be loaded, but data in DataFrame will not be cut off.
            Their length is limited only after calling :meth:`~_dict_to_interaction` or
            :meth:`~_dataframe_to_interaction`
        """
        self.logger.debug(set_color(f'Loading feature from [{filepath}] (source: [{source}]).', 'green'))

        load_col, unload_col = self._get_load_and_unload_col(source)
        if load_col == set():
            return None

        field_separator = self.config['field_separator']
        columns = []
        usecols = []
        dtype = {}
        encoding = self.config['encoding']
        with open(filepath, 'r', encoding=encoding) as f:
            head = f.readline()[:-1]
        for field_type in head.split(field_separator):
            field, ftype = field_type.split(':')
            try:
                ftype = FeatureType(ftype)
            except ValueError:
                raise ValueError(f'Type {ftype} from field {field} is not supported.')
            if load_col is not None and field not in load_col:
                continue
            if unload_col is not None and field in unload_col:
                continue
            if isinstance(source, FeatureSource) or source != 'link':
                self.field2source[field] = source
                self.field2type[field] = ftype
                if not ftype.value.endswith('seq'):
                    self.field2seqlen[field] = 1
            columns.append(field)
            usecols.append(field_type)
            dtype[field_type] = np.float64 if ftype == FeatureType.FLOAT else str

        if len(columns) == 0:
            self.logger.warning(f'No columns has been loaded from [{source}]')
            return None

        df: pd.DataFrame = self.perturb_feat(source)
        if df is None or source not in self.FEATS:
            df = pd.read_csv(
                filepath, delimiter=field_separator, usecols=usecols, dtype=dtype, encoding=encoding, engine='python'
            )
            df.columns = columns
            df = df.astype(dtype)

        seq_separator = self.config['seq_separator']
        for field in columns:
            ftype = self.field2type[field]
            if not ftype.value.endswith('seq'):
                continue
            df[field].fillna(value='', inplace=True)
            if ftype == FeatureType.TOKEN_SEQ:
                df[field] = [np.array(list(filter(None, _.split(seq_separator)))) for _ in df[field].values]
            elif ftype == FeatureType.FLOAT_SEQ:
                df[field] = [np.array(list(map(float, filter(None, _.split(seq_separator))))) for _ in df[field].values]
            self.field2seqlen[field] = max(map(len, df[field].values))
        return df



import os
import shutil
import decimal

import tqdm
import pandas as pd
from pandas.api.types import is_numeric_dtype


# def split_data_temporally(interactions,
#                           start_train_split=0.80,
#                           test_split=None,
#                           validation_split=None,
#                           user_field='user_id',
#                           time_field='timestamp'):
#     # TODO: each temporal split could be yielded instead of returning an entire structure with all the splits
#     # TODO: funzione che trova il set di giorni più vicini al N% (e.g. 10%, 20%), in modo che i giorni siano un numero
#     #       tondo, tipo 90 giorni (3 mesi) o 100
#     # TODO: trovare un modo efficiente per controllare se un timestamp sta in un intervallo di tempo
#     train_set = []
#     test_set = []
#     val_set = []
#     groups = interactions.groupby([user_field])
#     for i, (_, group) in enumerate(tqdm.tqdm(groups, desc=f"Splitting data per_user")):
#         if time_field:
#             sorted_group = group.sort_values(time_field)
#         else:
#             sorted_group = group
#
#         if isinstance(train_split, float) or isinstance(test_split, float):
#             n_rating_train = int(len(sorted_group.index) * train_split) if train_split is not None else 0
#             n_rating_test = int(len(sorted_group.index) * test_split) if test_split is not None else 0
#             n_rating_val = int(len(sorted_group.index) * validation_split) if validation_split is not None else 0
#
#             if len(sorted_group.index) > (n_rating_train + n_rating_test + n_rating_val):
#                 n_rating_train += len(sorted_group.index) - (n_rating_train + n_rating_test + n_rating_val)
#         else:
#             raise ValueError(f"split type not accepted")
#
#         if n_rating_train == 0:
#             start_index = len(sorted_group) - n_rating_test
#             start_index = start_index - n_rating_val if n_rating_val is not None else start_index
#             train_set.append(sorted_group.iloc[:start_index])
#         else:
#             train_set.append(sorted_group.iloc[:n_rating_train])
#             start_index = n_rating_train
#
#         if n_rating_val > 0:
#             val_set.append(sorted_group.iloc[start_index:(start_index + n_rating_val)])
#             start_index += n_rating_val
#
#         if n_rating_test > 0:
#             test_set.append(sorted_group.iloc[start_index:(start_index + n_rating_test)])
#         else:
#             test_set.append(sorted_group.iloc[start_index:])
#
#     train, test = pd.concat(train_set), pd.concat(test_set)
#     validation = pd.concat(val_set) if val_set else None
#
#     return train, validation, test


def split_data_per_user(interactions,
                        train_split=0.80,
                        test_split=None,
                        validation_split=None,
                        user_field='user_id',
                        time_field='timestamp'):
    train_set = []
    test_set = []
    val_set = []
    groups = interactions.groupby([user_field])
    for i, (_, group) in enumerate(tqdm.tqdm(groups, desc=f"Splitting data per_user")):
        if time_field in group.columns:
            sorted_group = group.sort_values(time_field)
        else:
            sorted_group = group

        # if train_split is a tuple the last train_split[1] interactions in the train are moved to the test
        if isinstance(train_split, tuple):
            tr_split, tr_to_test = train_split
            if not isinstance(tr_to_test, int):
                raise ValueError("the number of interactions to move from train to test must be an integer")
            n_train, n_val, n_test = _compute_set_sizes(len(sorted_group), (tr_split, validation_split, test_split))
            tr_split = n_train - tr_to_test
            te_split = n_test + tr_to_test
            val_split = n_val
        else:
            tr_split, val_split, te_split = train_split, validation_split, test_split

        _add_split_interactions(
            sorted_group,
            (tr_split, val_split, te_split),
            (train_set, val_set, test_set)
        )

    train, test = pd.concat(train_set), pd.concat(test_set)
    validation = pd.concat(val_set) if val_set else None

    return train, validation, test


def random_split(interactions: pd.DataFrame,
                 train_split=0.80,
                 test_split=None,
                 validation_split=None,
                 random_state=None):
    train_set = []
    test_set = []
    val_set = []

    interactions = interactions.sample(frac=1, random_state=random_state)

    _add_split_interactions(
        interactions,
        (train_split, validation_split, test_split),
        (train_set, val_set, test_set)
    )

    train, test = pd.concat(train_set), pd.concat(test_set)
    validation = pd.concat(val_set) if val_set else None

    return train, validation, test


def _add_split_interactions(dataset, splits, sets):
    train_set, val_set, test_set = sets
    if isinstance(splits[0], float):
        n_rating_train, n_rating_val, n_rating_test = _compute_set_sizes(len(dataset), splits)
    else:
        n_rating_train, n_rating_val, n_rating_test = splits

    if n_rating_train == 0:
        start_index = len(dataset) - n_rating_test
        start_index = start_index - n_rating_val if n_rating_val is not None else start_index
        train_set.append(dataset.iloc[:start_index])
    else:
        train_set.append(dataset.iloc[:n_rating_train])
        start_index = n_rating_train

    if n_rating_val > 0:
        val_set.append(dataset.iloc[start_index:(start_index + n_rating_val)])
        start_index += n_rating_val

    if n_rating_test > 0:
        test_set.append(dataset.iloc[start_index:(start_index + n_rating_test)])
    else:
        test_set.append(dataset.iloc[start_index:])


def _compute_set_sizes(n_data, splits):
    train_split, validation_split, test_split = splits

    if isinstance(train_split, float) or isinstance(test_split, float):
        n_rating_train = int(n_data * train_split) if train_split is not None else 0
        n_rating_val = int(n_data * validation_split) if validation_split is not None else 0
        n_rating_test = int(n_data * test_split) if test_split is not None else 0

        if n_data > (n_rating_train + n_rating_test + n_rating_val):
            n_rating_train += n_data - (n_rating_train + n_rating_test + n_rating_val)
    else:
        raise ValueError(f"split type not accepted")

    return n_rating_train, n_rating_val, n_rating_test


def filter_min_interactions(interactions, by='user_id', min_interactions=20):
    return interactions.groupby(by).filter(lambda x: len(x) >= min_interactions)


def copy_dataset(datasets_path, orig_dset_name, dset_new_name):
    dst = os.path.join(datasets_path, dset_new_name)
    shutil.copytree(os.path.join(datasets_path, orig_dset_name), dst, ignore=lambda x, y: ['splits_backup'])

    for dirname, _, filename in os.walk(dst, topdown=True):
        for data_type in recbole_extended_data_types:
            if f"{args.dataset}.{data_type}" == filename:
                os.rename(
                    os.path.join(dirname, filename),
                    os.path.join(dirname, f"{dset_new_name}.{data_type}")
                )


def add_token(df, user_df, args):
    for _df in [df, user_df]:
        new_cols = []
        for col in _df.columns:
            suffix = ':token'
            if is_numeric_dtype(_df[col]) and col not in [args.user_field, args.item_field]:
                suffix = ':float'
            new_cols.append(col + suffix)
        _df.columns = new_cols

    args.user_field += ':token'
    args.item_field += ':token'
    args.time_field += ':float'



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as mpl_tick


def legend_without_duplicate_labels(ax):
    handles, labels = ax.get_legend_handles_labels()
    unique = [(h, l) for i, (h, l) in enumerate(zip(handles, labels)) if l not in labels[:i]]
    ax.legend(*zip(*unique))
    return unique


def off_margin_ticks(*axs, axis='x'):
    # Turn off tick visibility for the measure axis on the marginal plots
    f = f"get_{axis}ticklabels"

    for ax in axs:
        plt.setp(getattr(ax, f)(), visible=False)
        plt.setp(getattr(ax, f)(minor=True), visible=False)


def make_patch_spines_invisible(ax):
    ax.set_frame_on(True)
    ax.patch.set_visible(False)
    for sp in ax.spines.values():
        sp.set_visible(False)


def add_bar_value_labels(ax, spacing=5, format='.2f', **kwargs):
    """Add labels to the end of each bar in a bar chart.
    https://stackoverflow.com/a/48372659
    Arguments:
        ax (matplotlib.axes.Axes): The matplotlib object containing the axes
            of the plot to annotate.
        spacing (int): The distance between the labels and the bars.
        format (str): format of the value of the bars.
    """

    # For each bar: Place a label
    for rect in ax.patches:
        # Get X and Y placement of label from rect.
        y_value = rect.get_height()
        x_value = rect.get_x() + rect.get_width() / 2

        # Number of points between bar and label. Change to your liking.
        space = spacing
        # Vertical alignment for positive values
        va = 'bottom'

        # If value of bar is negative: Place label below bar
        if y_value < 0:
            # Invert space to place label below
            space *= -1
            # Vertically align label at top
            va = 'top'

        # Use Y value as label and format number with one decimal place
        label = ("{:" + format + "}").format(y_value)

        # Create annotation
        ax.annotate(
            label,                       # Use `label` as label
            (x_value, y_value),          # Place label at end of the bar
            xytext=(0, space),           # Vertically shift label by `space`
            textcoords="offset points",  # Interpret `xytext` as offset in points
            ha='center',                 # Horizontally center label
            va=va,                       # Vertically align label differently for
            **kwargs)                    # positive and negative values.


def annotate_brackets(ax, num1, num2, data, center, height, yerr=None, dh=.05, barh=.05, same_h=True, fs=None, pval_th=0.001):
    """
    https://stackoverflow.com/a/52333561

    Annotate plot (barplot, boxplot, ...) with p-values.

    :param num1: number of left bar to put bracket over
    :param num2: number of right bar to put bracket over
    :param data: string to write or number for generating asterixes
    :param center: centers of all bars (like plt.bar() input)
    :param height: heights of all bars (like plt.bar() input)
    :param yerr: yerrs of all bars (like plt.bar() input)
    :param dh: height offset over bar / bar + yerr in axes coordinates (0 to 1)
    :param barh: bar height in axes coordinates (0 to 1)
    :param same_h: if the bracket edges should have the same length or not (if not the ends get close to the bars)
    :param fs: font size
    """

    text = f"p = {data:.3f}" if data >= pval_th else "p < " + str(pval_th)

    lx, ly = center[num1], height[num1]
    rx, ry = center[num2], height[num2]

    if yerr:
        ly += yerr[num1]
        ry += yerr[num2]

    dh *= max(ly, ry)
    barh *= max(ly, ry)

    if same_h:
        y = max(ly, ry) + dh
        y0, y1 = y, y

        barhl = {0: barh, 1: barh}
    else:
        y0 = ly + dh
        y1 = ry + dh

        barhl = dict.fromkeys([0, 1])
        if y0 > y1:
            barhl[1] = y0 + barh - y1
            barhl[0] = barh
        else:
            barhl[0] = y1 + barh - y0
            barhl[1] = barh

    barx = [lx, lx, rx, rx]
    bary = [y0, y0+barhl[0], y1+barhl[1], y1]
    mid = ((lx+rx)/2, max(y0, y1)+barh)

    ax.plot(barx, bary, c='black')
    # ax.plot([lx + 0.001, rx - 0.001], [max(y0, y1), max(y0, y1)], c='black', ls='--', lw=0.6)

    kwargs = dict(ha='center', va='bottom')
    if fs is not None:
        kwargs['fontsize'] = fs

    return ax.text(*mid, text, **kwargs)


def hierarchical_labels(ax, axis="x", tick_pos=None, label_sep='-', offset=-0.1):
    if axis == "x":
        twin = "twiny"
        _axis = "xaxis"
        spine = "bottom"
        ticks = "xticklabels"
        offset = -abs(offset)
    else:
        twin = "twinx"
        _axis = "yaxis"
        spine = "left"
        ticks = "yticklabels"
        offset = abs(offset)

    labels = getattr(ax, f"get_{ticks}")()
    levels = list(zip(*map(lambda x: x.get_text().split(label_sep), labels)))[::-1]
    tick_pos = [np.linspace(0, 1, len(l) + 1) for l in levels] if tick_pos is None else tick_pos
    for level, (level_labels, tick_p) in enumerate(zip(levels, tick_pos)):
        twin_ax = getattr(ax, twin)() if level > 0 else ax

        twin_ax.spines[spine].set_position(("axes", offset * (level + 1)))
        twin_ax.tick_params('both', length=0, width=0, which='minor')
        twin_ax.tick_params('both', direction='in', which='major')
        getattr(twin_ax, _axis).set_ticks_position(spine)
        getattr(twin_ax, _axis).set_label_position(spine)

        twin_ax.set_xticks(tick_p)
        getattr(twin_ax, _axis).set_major_formatter(mpl_tick.NullFormatter())
        loc = [np.mean(tick_p[i - 1:i + 1]) for i in range(1, len(tick_p))]
        getattr(twin_ax, _axis).set_minor_locator(mpl_tick.FixedLocator(loc))
        getattr(twin_ax, _axis).set_minor_formatter(mpl_tick.FixedFormatter(level_labels))



# -*- coding: utf-8 -*-

import sys

import torch
import torch.nn as nn
import torch.nn.functional as F

sys.path.append('..')

from .ngcf import NGCF
from gnnuers.models import PerturbedModel


class NGCFPerturbed(PerturbedModel, NGCF):
    def __init__(self, config, dataset, **kwargs):
        NGCF.__init__(self, config, dataset)
        PerturbedModel.__init__(self, config, **kwargs)

    def forward(self, pred=False):
        A_hat = self.perturb_adj_matrix(self.Graph, pred=pred)
        if self.node_dropout != 0:
            A_hat = self.sparse_dropout(A_hat)

        all_embeddings = self.get_ego_embeddings()
        embeddings_list = [all_embeddings]
        for gnn in self.GNNlayers:
            all_embeddings = gnn(A_hat, self.eye_matrix, all_embeddings)
            all_embeddings = nn.LeakyReLU(negative_slope=0.2)(all_embeddings)
            all_embeddings = self.emb_dropout(all_embeddings)
            all_embeddings = F.normalize(all_embeddings, p=2, dim=1)
            embeddings_list += [all_embeddings]  # storage output embedding of each layer
        ngcf_all_embeddings = torch.cat(embeddings_list, dim=1)

        user_all_embeddings, item_all_embeddings = torch.split(ngcf_all_embeddings, [self.n_users, self.n_items])
        return user_all_embeddings, item_all_embeddings



import os
import argparse
import pickle as pkl
from datetime import datetime

import tqdm
import pandas as pd  # better use pandas >= 2.0.0
import numpy as np


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--folderpath', required=True)
    parser.add_argument('--chunksize', default=50_000_000)
    parser.add_argument('--min_interactions', default=1_000)
    parser.add_argument('--overwrite', action="store_true")

    args = parser.parse_args()

    le_size = 2_014_164_872
    lc_size = 519_293_333

    pp_users_filepath = os.path.join(args.folderpath, 'preprocessed_lastfm2b_users.tsv')
    if args.overwrite or not os.path.exists(pp_users_filepath):
        users_df = pd.read_csv(os.path.join(args.folderpath, 'users.tsv'), sep='\t', on_bad_lines='warn')

        age_mask = (users_df.age > 0) & (users_df.age < 120)
        gender_mask = (users_df.gender == 'm') | (users_df.gender == 'f')
        users_df = users_df[age_mask & gender_mask]
        users_df['gender'] = users_df.gender.str.upper()
        users_df.to_csv(pp_users_filepath, sep='\t', index=None)
    else:
        users_df = pd.read_csv(pp_users_filepath, sep='\t')

    tracks_albums_filepath = os.path.join(args.folderpath, 'tracks_albums_df.tsv')
    if args.overwrite or not os.path.exists(tracks_albums_filepath):
        tracks_df = pd.read_csv(
            os.path.join(args.folderpath, 'tracks.tsv'),
            usecols=['track_id', 'artist_name'],
            sep='\t', on_bad_lines='warn'
        )
        albums_df = pd.read_csv(
            os.path.join(args.folderpath, 'albums.tsv'),
            usecols=['album_id', 'artist_name'],
            sep='\t', on_bad_lines='warn'
        )

        tracks_albums_ids_filepath = os.path.join(args.folderpath, 'tracks_albums_ids.npy')
        if args.overwrite or not os.path.exists(tracks_albums_ids_filepath):
            tracks_albums_ids = None
            le_kwargs = dict(
                usecols=['track_id', 'album_id'],
                sep='\t',
                on_bad_lines='warn',
                chunksize=args.chunksize
            )
            with pd.read_csv(os.path.join(args.folderpath, 'listening-events.tsv'), **le_kwargs) as reader:
                for chunk in tqdm.tqdm(
                    reader, total=int(np.ceil(le_size / args.chunksize)),
                    desc='Extracting the tracks and albums ids to identify unique artists'
                ):
                    if tracks_albums_ids is None:
                        tracks_albums_ids = np.unique(chunk.values, axis=0)
                    else:
                        tracks_albums_ids = np.unique(
                            np.concatenate((tracks_albums_ids, np.unique(chunk.values, axis=0)), axis=0),
                            axis=0
                        )
            if tracks_albums_ids is not None:
                np.save(tracks_albums_ids_filepath, tracks_albums_ids)
        else:
            tracks_albums_ids = np.load(tracks_albums_ids_filepath)

        # import pdb; pdb.set_trace()
        tr_alb_df = pd.DataFrame(tracks_albums_ids, columns=['track_id', 'album_id'])
        tr_alb_df = tr_alb_df.join(tracks_df.set_index('track_id'), on='track_id')
        tr_alb_df = tr_alb_df.join(albums_df.set_index('album_id'), on='album_id', rsuffix='_alb')

        an_na = tr_alb_df['artist_name'].isna() & ~tr_alb_df['artist_name_alb'].isna()
        an_alb_na = ~tr_alb_df['artist_name'].isna() & tr_alb_df['artist_name_alb'].isna()

        tr_alb_df.loc[an_na, 'artist_name'] = tr_alb_df.loc[an_na, 'artist_name_alb']
        tr_alb_df.loc[an_alb_na, 'artist_name_alb'] = tr_alb_df.loc[an_alb_na, 'artist_name']

        tr_alb_df = tr_alb_df[tr_alb_df.artist_name == tr_alb_df.artist_name_alb]  # removes also NaN
        del tr_alb_df['artist_name_alb']

        an_unique = tr_alb_df['artist_name'].unique()
        an_aid_map = dict(zip(an_unique, np.arange(an_unique.shape[0])))
        tr_alb_df['artist_id'] = tr_alb_df['artist_name'].map(an_aid_map)

        # some tracks can be in multiple albums
        # now we delete some tracks with equal ids but different artist_names
        tr_aid = tr_alb_df[['track_id', 'artist_id']].values
        tr_aid_unq = np.unique(tr_aid, axis=0)
        tr_unq, tr_cnt = np.unique(tr_aid_unq[:, 0], return_counts=True)
        bad_tracks = tr_unq[tr_cnt > 1]

        del tr_alb_df['artist_id']

        tr_alb_df = tr_alb_df[~tr_alb_df.track_id.isin(bad_tracks)]

        tr_alb_df.to_csv(tracks_albums_filepath, sep='\t', index=None)
    else:
        tr_alb_df = pd.read_csv(tracks_albums_filepath, sep='\t')

    track_artist_map = dict(tr_alb_df[['track_id', 'artist_name']].values)
    del tr_alb_df

    ua_df_filepath = os.path.join(args.folderpath, 'user_artist_df.tsv')
    if args.overwrite or not os.path.exists(ua_df_filepath):
        user_artist_count = []
        lc_kwargs = dict(
            sep='\t',
            on_bad_lines='warn',
            chunksize=args.chunksize
        )
        with pd.read_csv(os.path.join(args.folderpath, 'listening-counts.tsv'), **lc_kwargs) as reader:
            for chunk in tqdm.tqdm(
                reader, total=int(np.ceil(lc_size / args.chunksize)),
                desc='Extracting the listening counts for each track and taking the sum grouping by artist name'
            ):
                chunk['artist_name'] = chunk['track_id'].map(lambda x: track_artist_map.get(x, None))
                chunk.dropna(inplace=True)
                del chunk['track_id']
                user_artist_count.append(
                    chunk.groupby(['user_id', 'artist_name']).sum().reset_index()
                )

        # import pdb; pdb.set_trace()
        ua_df = pd.concat(user_artist_count, ignore_index=True)
        ua_df = ua_df.groupby(['user_id', 'artist_name']).sum().reset_index()

        ua_df.to_csv(ua_df_filepath, sep='\t', index=None)
    else:
        ua_df = pd.read_csv(ua_df_filepath, sep='\t')

    # only users with gender and age
    ua_df = ua_df[ua_df.user_id.isin(users_df.user_id)]
    ua_df_idx = ua_df.set_index(['user_id', 'artist_name'])

    ua_last_timestamp_path = os.path.join(args.folderpath, 'user_artist_last_timestamp.npy')
    if args.overwrite or not os.path.exists(ua_last_timestamp_df_path):
        unique_ua_pos = dict(zip(ua_df_idx.index.tolist(), np.arange(len(ua_df_idx))))
        ua_last_timestamp = np.full((len(ua_df_idx),), -1, dtype=int)
        le_kwargs = dict(
            usecols=['user_id', 'track_id', 'timestamp'],
            sep='\t',
            on_bad_lines='warn',
            chunksize=args.chunksize
        )
        with pd.read_csv(os.path.join(args.folderpath, 'listening-events.tsv'), **le_kwargs) as reader:
            for chunk in tqdm.tqdm(
                reader, total=int(np.ceil(le_size / args.chunksize)),
                desc='Retrieving the last timestamp each artist (the track of the artist) was listened to'
            ):
                chunk['timestamp'] = chunk['timestamp'].map(lambda x: int(datetime.fromisoformat(x).timestamp()))
                chunk['artist_name'] = chunk['track_id'].map(lambda x: track_artist_map.get(x, None))
                chunk.dropna(inplace=True)
                del chunk['track_id']
                chunk_time = chunk.groupby(['user_id', 'artist_name']).max().to_dict()['timestamp']
                for k, v in chunk_time.items():
                    if k in unique_ua_pos:
                        ua_last_timestamp[unique_ua_pos[k]] = max(unique_ua_pos[k], v)

        del unique_ua_pos
        np.save(ua_last_timestamp_path, ua_last_timestamp)
    else:
        ua_last_timestamp = np.load(ua_last_timestamp_path)

    ua_df_idx['timestamp'] = ua_last_timestamp
    ua_df_idx.to_csv(os.path.join(args.folderpath, 'preprocessed_lastfm2b_inter.tsv'), sep='\t')

    ua_df_1000 = ua_df_idx.reset_index().groupby('user_id').filter(lambda x: len(x) >= args.min_interactions)
    ua_df_idx.to_csv(os.path.join(args.folderpath, 'preprocessed_lastfm2b_inter_1000.tsv'), sep='\t', index=None)

    ua_df_idx.to_csv(os.path.join(args.folderpath, 'preprocessed_lastfm2b_users_1000.tsv'), sep='\t', index=None)
    users_df = users_df[users_df.user_id.isin(ua_df_idx.user_id.unique())]
    users_df.to_csv(os.path.join(args.folderpath, 'preprocessed_lastfm2b_users_1000.tsv'), sep='\t', index=None)



# -*- coding: utf-8 -*-

import sys

import torch
from recbole.model.general_recommender.gcmc import GCMC, GcEncoder

sys.path.append('..')

from gnnuers.models import PerturbedModel


class GCMCPerturbed(PerturbedModel, GCMC):
    def __init__(self, config, dataset, **kwargs):
        GCMC.__init__(self, config, dataset)
        PerturbedModel.__init__(self, config, **kwargs)

        self.support[0] = self.Graph

        self.GcEncoder = GcEncoderPerturbed(
            accum=self.accum,
            num_user=self.n_users,
            num_item=self.n_items,
            support=self.support,
            input_dim=self.input_dim,
            gcn_output_dim=self.gcn_output_dim,
            dense_output_dim=self.dense_output_dim,
            drop_prob=self.dropout_prob,
            device=self.device,
            edge_additions=self.edge_additions,
            mask_sub_adj=self.mask_sub_adj,
            only_subgraph=self.only_subgraph,
            force_removed_edges=self.force_removed_edges,
            perturb_adj_matrix_func=self.perturb_adj_matrix,
            sparse_feature=self.sparse_feature
        ).to(self.device)

    def forward(self, user_X, item_X, user, item, pred=False):
        # Graph autoencoders are comprised of a graph encoder model and a pairwise decoder model.
        user_embedding, item_embedding = self.GcEncoder(user_X, item_X, pred=pred)
        predict_score = self.BiDecoder(user_embedding, item_embedding, user, item)
        return predict_score

    def predict(self, interaction, pred=False):
        user = interaction[self.USER_ID]
        item = interaction[self.ITEM_ID]

        user_X, item_X = self.user_features, self.item_features
        predict = self.forward(user_X, item_X, user, item, pred=pred)

        score = predict[:, 1]
        return score

    def full_sort_predict(self, interaction, pred=False):
        user = interaction[self.USER_ID]

        user_X, item_X = self.user_features, self.item_features
        predict = self.forward(user_X, item_X, user, None, pred=pred)

        score = predict[:, 1]
        return score


class GcEncoderPerturbed(GcEncoder):

    def __init__(
        self,
        accum,
        num_user,
        num_item,
        support,
        input_dim,
        gcn_output_dim,
        dense_output_dim,
        drop_prob,
        device,
        edge_additions,
        mask_sub_adj,
        only_subgraph,
        force_removed_edges,
        perturb_adj_matrix_func,
        sparse_feature=True,
        act_dense=lambda x: x,
        share_user_item_weights=True,
        bias=False
    ):
        super(GcEncoderPerturbed, self).__init__(
            accum=accum,
            num_user=num_user,
            num_item=num_item,
            support=support,
            input_dim=input_dim,
            gcn_output_dim=gcn_output_dim,
            dense_output_dim=dense_output_dim,
            drop_prob=drop_prob,
            device=device,
            sparse_feature=sparse_feature,
            act_dense=act_dense,
            share_user_item_weights=share_user_item_weights,
            bias=bias
        )
        self.num_all = self.num_users + self.num_items

        self.edge_additions = edge_additions
        self.P_hat_symm, self.P = None, None

        self.mask_sub_adj = mask_sub_adj

        self.D_indices = torch.arange(self.num_all).tile((2, 1)).to(self.device)
        self.only_subgraph = only_subgraph

        self.force_removed_edges = force_removed_edges

        self.perturb_adj_matrix = perturb_adj_matrix_func

    def forward(self, user_X, item_X, pred=False):
        # ----------------------------------------GCN layer----------------------------------------

        user_X = self.sparse_dropout(user_X)
        item_X = self.sparse_dropout(item_X)

        embeddings = []
        if self.accum == 'sum':
            wu = 0.
            wv = 0.
            for i in range(self.num_support):
                # weight sharing
                wu = self.weights_u[i] + wu
                wv = self.weights_v[i] + wv

                # multiply feature matrices with weights
                if self.sparse_feature:
                    temp_u = torch.sparse.mm(user_X, wu)
                    temp_v = torch.sparse.mm(item_X, wv)
                else:
                    temp_u = torch.mm(user_X, wu)
                    temp_v = torch.mm(item_X, wv)
                all_embedding = torch.cat([temp_u, temp_v])

                norm_adj = self.perturb_adj_matrix(self.support[i], pred=pred)
                all_emb = torch.sparse.mm(norm_adj, all_embedding)
                embeddings.append(all_emb)

            embeddings = torch.stack(embeddings, dim=1)
            embeddings = torch.sum(embeddings, dim=1)
        else:
            for i in range(self.num_support):
                # multiply feature matrices with weights
                if self.sparse_feature:
                    temp_u = torch.sparse.mm(user_X, self.weights_u[i])
                    temp_v = torch.sparse.mm(item_X, self.weights_v[i])
                else:
                    temp_u = torch.mm(user_X, self.weights_u[i])
                    temp_v = torch.mm(item_X, self.weights_v[i])
                all_embedding = torch.cat([temp_u, temp_v])

                # then multiply with adj matrices
                norm_adj = self.perturb_adj_matrix(self.support[i], pred=pred)
                all_emb = torch.sparse.mm(norm_adj, all_embedding)
                embeddings.append(all_emb)

            embeddings = torch.cat(embeddings, dim=1)

        users, items = torch.split(embeddings, [self.num_users, self.num_items])

        u_hidden = self.activate(users)
        v_hidden = self.activate(items)

        # ----------------------------------------Dense Layer----------------------------------------

        u_hidden = self.dropout(u_hidden)
        v_hidden = self.dropout(v_hidden)

        u_hidden = self.dense_layer_u(u_hidden)
        v_hidden = self.dense_layer_u(v_hidden)

        u_outputs = self.dense_activate(u_hidden)
        v_outputs = self.dense_activate(v_hidden)

        return u_outputs, v_outputs



from .ranking_losses import NDCGApproxLoss, SigmoidBCELoss
from .beyondacc_losses import ConsumerDPLoss, ProviderDPLoss, UserCoverageLoss


def get_ranking_loss(loss="ndcg"):
    return {
        "ndcg": NDCGApproxLoss,
        "sigmoid_bce": SigmoidBCELoss
    }[loss.lower()]


def get_beyondacc_loss(loss="consumer_dp"):
    return {
        "consumer_dp": ConsumerDPLoss,
        "provider_dp": ProviderDPLoss,
        "uc": UserCoverageLoss
    }[loss.lower()]


__exp_metrics_to_losses_map__ = {
    "consumer_DP": ConsumerDPLoss,
    "consumer_DP_across_random_samples": ConsumerDPLoss,
    "UC": UserCoverageLoss,
    "provider_DP": ProviderDPLoss
}


def get_loss_from_exp_metric(exp_metric):
    if exp_metric in __exp_metrics_to_losses_map__:
        return __exp_metrics_to_losses_map__[exp_metric]
    else:
        raise NotImplementedError(f'The exp_metric `{exp_metric}` is not supported')



import os
from setuptools import setup, find_packages

try:
    import torch
except ModuleNotFoundError:
    raise ModuleNotFoundError("torch must be installed first")

try:
    import torch_geometric
except ModuleNotFoundError:
    raise ModuleNotFoundError("torch-geometric must be installed first")

try:
    import torch_scatter
except ModuleNotFoundError:
    raise ModuleNotFoundError("torch-scatter must be installed first")

try:
    import torch_sparse
except ModuleNotFoundError:
    raise ModuleNotFoundError("torch-sparse must be installed first")

install_requires = [
    "recbole>=1.2.0",
    "numba",
    "wandb",
    "igraph",
    "optuna",
    "seaborn>=0.11.2"
]

setup_requires = []

extras_require = {}

classifiers = [
    "Development Status :: 3 - Alpha",
    "Environment :: GPU :: NVIDIA CUDA",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
]

long_description = """
GNNUERS generates explanations in the form of user-item interactions that make
a GNN-based recommender system favor a demographic group over another. GNNUERS learns a perturbation
vector that modifies the adjacency matrix representing the training network. The edges modified by
the perturbation vector are the explanations genereated by the framework. GNNUERS then needs to work on a slight
extended version of a recommender system in order to include the perturbation vector and optimize it.
"""

setup(
    name="gnnuers",
    version="0.0.3.1",  # please remember to edit gnnuers/__init__.py in response, once updating the version
    description="A library to generate explanations of recommendations generated by GNNs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jackmedda/RS-BGExplainer",
    author="jackmedda",
    author_email="jackm.medda@gmail.com",
    packages=[package for package in find_packages() if package.startswith("gnnuers")],
    include_package_data=True,
    install_requires=install_requires,
    setup_requires=setup_requires,
    extras_require=extras_require,
    zip_safe=False,
    classifiers=classifiers,
)



import numpy as np
import scipy.signal as sp_signal


class EarlyStopping:
    def __init__(self, patience, ignore, method='consistency', delta=1e-4, fn='lt', **kwargs):
        '''
        Args:
            patience : Integer, number of previous loss values stored at one time
            ignore : Integer, number of epochs to ignore
            method : String, 'consistency' breaks the training loop if the stored loss values are strictly increasing. 'slope' breaks the training loop if the regression line for the loss values equals or exceeds 0.
        '''
        self.patience = patience
        self.ignore = ignore
        self.method = method
        self.delta = delta
        self.fn = f'__{fn}__'
        self.history = []
        self.best_loss = 0
        self.methods = {
            'slope': self.check_slope,
            'consistency': self.check_consistency,
            'savgol': self.check_savinsky_golay_consistency
        }

        self._min_epochs = kwargs.get('min_epochs', 10)
        self._window_length = kwargs.get('window_length', None)
        if self._window_length is None:
            self._window_length = self._deafult_window_length
        self._polyorder = kwargs.get('polyorder', 2)

    def _deafult_window_length(x):
        return len(x) // 2

    def check(self, loss):
        '''
        Determines whether or not to break the training loop given a new loss value.

        Args:
            loss : Integer/Float, loss value to be considered

        Returns:
            Boolean : True means that the training loop should be broken
        '''
        self.history.append(loss)

        if self.is_stoppable():
            return True

        return False

    def is_stoppable(self):
        if len(self.history) > self.ignore:
            return self.methods[self.method]()

        self.best_loss = len(self.history) - 1
        return False

    def check_slope(self):
        raise NotImplementedError()
        x = [i for i in range(len(self.queue))]
        n = len(x)
        y = self.queue
        return (n*(sum([x[i]*y[i] for i in range(n)])) - sum(x)*sum(y))/(n*sum([v**2 for v in x]) - sum(x)**2) >= 0

    def check_consistency(self):
        last = self.history[-1]
        if getattr(last, self.fn)(self.history[self.best_loss]) and abs(last - self.history[self.best_loss]) > self.delta:
            self.best_loss = len(self.history) - 1
            return False
        return len(self.history) - 1 - self.best_loss >= self.patience

    def check_savinsky_golay_consistency(self):
        if len(self.history) >= self._min_epochs:
            history_smoothed = sp_signal.savgol_filter(self.history, self._window_length(self.history), self._polyorder)
            last = history_smoothed[-1]
            if getattr(last, self.fn)(history_smoothed[self.best_loss]) and abs(last - history_smoothed[self.best_loss]) > self.delta:
                self.best_loss = np.argmin(history_smoothed)
                return False
        return len(self.history) - 1 - self.best_loss >= self.patience

    def reset(self):
        '''
        Resets the EarlyStopping object
        '''
        self.__init__(self.patience, self.ignore, self.method, self.fn)

    def __str__(self):
        return f"History: {str(self.history)}; Best loss history index: {self.best_loss}"

    def __repr__(self):
        return self.__str__()



# %%
import os
import argparse
import inspect

import pandas as pd

import gnnuers.utils as utils


# %%
def get_plots_path():
    plots_path = os.path.join(
        script_path,
        f'new_dp_plots'
    )

    if not os.path.exists(plots_path):
        os.makedirs(plots_path)

    return plots_path


# %%
parser = argparse.ArgumentParser()
parser.add_argument('--saved_path', default='saved')

args = parser.parse_args()

script_path = os.path.abspath(os.path.dirname(inspect.getsourcefile(lambda: 0)))
if os.sep not in args.saved_path:
    args.saved_path = os.path.join(script_path, os.pardir, args.saved_path)

config_path = os.path.join(args.saved_path, os.pardir, 'config')
base_conf_exp_name = 'base_explainer.yaml'
yaml_loader = Config('GCMC', 'ml-1m').yaml_loader

print(args)

check_datasets = []
out_data = []
out_columns = [
    '# Users',
    '# Items',
    '# Interactions',
    'Gender Repr.',
    'Age Repr.',
    'Min. Degree per user',
]

stats_path = os.path.join(get_plots_path(), 'datasets_stats.csv')
if not os.path.exists(stats_path):
    for model_path in os.listdir(args.saved_path):
        if 'perturbed' not in model_path and os.path.isfile(os.path.join(args.saved_path, model_path)):
            dset_name = model_path.replace('GCMC-', '').replace('LightGCN-', '').replace('NGCF-', '')  # removes model name
            dset_name = dset_name[::-1].split('_', maxsplit=1)[1].split('-', maxsplit=3)[-1][::-1]  # removes date
            if dset_name.lower() not in check_datasets:
                config_files = os.listdir(config_path)

                exp_conf_name = f"{dset_name.lower()}_explainer.yaml"
                if exp_conf_name in config_files:
                    exp_conf_file = os.path.join(config_path, exp_conf_name)
                else:
                    print(f'Using the base config explainer file. \'{dset_name.lower()}\'_explainer.yaml not found')
                    exp_conf_file = None

                config_dict = utils.update_base_explainer(
                    os.path.join(config_path, base_conf_exp_name),
                    exp_conf_file
                )

                check_datasets.append(dset_name.lower())
                config, model, dataset, train_data, valid_data, test_data = utils.load_data_and_model(
                    os.path.join(args.saved_path, model_path),
                    explainer_config=config_dict
                )

                user_feat = pd.DataFrame(dataset.user_feat.numpy()).iloc[1:]
                sens_cols = user_feat.columns[user_feat.columns.isin(['gender', 'age'])]
                sens_info = dict.fromkeys(['Gender', 'Age'])
                for col in sens_cols:
                    if col in dataset.field2id_token and col in ['age', 'gender']:
                        user_feat[col] = user_feat[col].map(dataset.field2id_token[col].__getitem__)
                        if col == 'age':
                            user_feat[col] = user_feat[col].map({'M': 'Y', 'F': 'O'})
                        col_info = (user_feat[[col]].value_counts() / len(user_feat) * 100).map(lambda x: f"{x:.1f}%")
                        sens_info[col.title()] = '; '.join(
                            col_info.to_frame().reset_index().apply(
                                lambda x: f"{x[col]} : {x[0]}".replace('%', '\%'), axis=1
                            ).values
                        )
                    else:
                        sens_info[col] = 'NA'

                out_data.append([
                    dataset.user_num - 1,
                    dataset.item_num - 1,
                    dataset.inter_num,
                    sens_info['Gender'],
                    sens_info['Age'],
                    dataset.history_item_matrix()[2][1:].min().item()
                ])

    df = pd.DataFrame(out_data, columns=out_columns, index=check_datasets)
    df.index.name = "Dataset"
    df = df.T
    print(df)
    df.to_csv(stats_path)
else:
    df = pd.read_csv(stats_path, index_col=0)


def dict_mapper(x):
    if isinstance(x, str):
        if 'Gender' in x or 'Age' in x:
            x = x.replace('Age ', '')
            return x.split(' Gender ') if 'Gender' in x else [x, '']
    return x

df = df[["ml-1m", "tafeng", "lastfm-1k", "insurance"]]
df.columns = ["ML-1M", "FENG", "LFM-1K", "INS"]
df.loc["Age Repr."] = df.loc["Age Repr."].str.replace('M :', 'Y :').str.replace('F :', 'O :')
df.loc["Domain"] = ["Movie", "Grocery", "Music", "Insurance"]
df.rename(index={
    "bar{Degree}": "$\overline{\text{User DEG}}$",
    "bar{Density}": "$\overline{\text{User DY}}$",
    "bar{Reachability}": "$\overline{\text{User IDG}}$",
    "Gini Degree": "Gini User DEG",
    "Gini Density": "Gini User DY",
    "Gini Reachability": "Gini User IDG",
    "Min. Degree per user": "Min. User DEG"
}, inplace=True)

df = df.applymap(dict_mapper).explode(df.columns.tolist())
df['Group'] = df["ML-1M"].map(
    lambda x: None if not isinstance(x, str) else ("G" if "F :" in x else ("A" if "O :" in x else None))
)
df = df.reset_index().set_index(['index', 'Group'])

with pd.option_context('max_colwidth', None):
    tex_table = df.to_latex(column_format="rr|r|r|r|r", escape=False, multirow=True)

with open(os.path.join(get_plots_path(), 'datasets_stats.tex'), 'w') as tex_f:
    tex_f.write(
        tex_table.replace(
            '# Users & NaN ', r'\multicolumn{2}{r|}{\# Users}'
        ).replace(
            '# Items & NaN ', r'\multicolumn{2}{r|}{\# Items}'
        ).replace(
            '# Interactions & NaN ', r'\multicolumn{2}{r|}{\# Interactions}'
        ).replace(
            'Min. User DEG & NaN ', r'\multicolumn{2}{r|}{Min. User DEG}'
        ).replace(
            'Domain & NaN ', r'\multicolumn{2}{r|}{Domain}'
        ).replace(
            'ML-1M', r'ML-1M \cite{DBLP:journals/tiis/HarperK16}'
        ).replace(
            'FENG', r'FENG\footnote{\url{https://www.kaggle.com/datasets/chiranjivdas09/ta-feng-grocery-dataset}}'
        ).replace(
            'LFM-1K', r'LFM-1K \cite{DBLP:books/daglib/0025137}'
        ).replace(
            'INS', r'INS\footnote{\url{https://www.kaggle.com/datasets/mrmorj/insurance-recommendation}}'
        ).replace(
            r'\cline{1-6}', r'\cline{2-6}'
        ).replace(
            'Gender Repr.', r'\multirow{2}{*}{Repr.}'
        ).replace(
            'Age Repr.', ''
        ).replace('NaN', 'NA')
    )



import os
import pprint
import argparse

import pandas as pd

from gnnuers.utils import copytree


def map_multi2binary(df, field, label):
    mask = df[field] == label
    df.loc[mask, field] = "M"
    df.loc[~mask, field] = "F"
    return df


if __name__ == "__main__":
    r"""
    python -m src.data.map_multi2binary --user_filepath dataset.user --map_attributes age
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset', '--d', required=True)
    parser.add_argument('--map_attributes', '--ma', nargs='+', required=True)
    parser.add_argument('--user_filepath', '--uf', default='')

    args = parser.parse_args()

    datasets_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), os.pardir, os.pardir, 'dataset'
    )

    uf = args.user_filepath
    uf = os.path.join(datasets_path, args.dataset, f"{args.dataset}.user") if not uf else uf
    user_df = pd.read_csv(uf, sep='\t')

    pprint.pprint(vars(args))

    recbole_extended_data_types = ['.inter', '.item', '.test', '.train', '.user', '.validation']

    for attr in args.map_attributes:
        unique_labels = user_df[attr].unique()
        if len(unique_labels) == 2:
            raise ValueError(f"The attribute `{attr}` is already binary")

        for label in unique_labels:
            df = user_df.copy(deep=True)
            df = map_multi2binary(df, attr, label)

            dset_new_name = f"{args.dataset}_{attr.split(':')[0] if ':' in attr else attr}_{label}"
            copy_dataset(datasets_path, args.dataset, dset_new_name)

            df.to_csv(os.path.join(dst, f"{args.dataset}.user"), index=None, sep='\t')

            print('-' * 50)
            print(f"Attr: {attr}    Label: {label}")
            print(df[attr].value_counts().map(lambda x: f"{x / len(df) * 100:.2f}%"))
